#ifndef __GEOM2_H__
#define __GEOM2_H__

#include <math.h>
#include <float.h>
#include <iostream>
#include <utility>

class vec2 {
 public:
  vec2() {
    _v[0] = 0;
    _v[1] = 0;
    _v[2] = 0;
  }
  vec2(float x, float y) {
    _v[0] = x;
    _v[1] = y;
    _v[2] = 0;
  }
  float* get() {
    return _v;
  }
  const float* get() const {
    return _v;
  }
  float x() const { return _v[0]; }
  float y() const { return _v[1]; }
  float length() const {
    return sqrt(x()*x()+y()*y());
  }
  vec2 normalized() const {
    return (*this)/length();
  }

  float dot(const vec2& v) const {
    return x()*v.x() + y()*v.y();
  }

  vec2 operator+(const vec2& v) const {
    return vec2(x()+v.x(), y()+v.y());
  }
  vec2 operator-(const vec2& v) const {
    return (*this)+v*(-1);
  }
  vec2 operator*(const float& f) const {
    return vec2(f*x(), f*y());
  }
  vec2 operator/(const float& f) const {
    return (*this)*(1/f);
  }

 private:
  float _v[3];
};

class point2 {
 public:
  point2() {
    _v[0] = 0;
    _v[1] = 0;
    _v[2] = 1;
  }
  point2(float x, float y) {
    _v[0] = x;
    _v[1] = y;
    _v[2] = 1;
  }
  float* get() {
    return _v;
  }
  const float* get() const {
    return _v;
  }
  float x() const { return _v[0]; }
  float y() const { return _v[1]; }

  vec2 operator-(const point2& v) const {
    return vec2(x()-v.x(), y()-v.y());
  }

  point2 operator+(const vec2& v) const {
    return point2(x()+v.x(), y()+v.y());
  }

 private:
  float _v[3];
};

inline vec2 operator*(const float f, const vec2& v) {
  return v * f;
}

// distance from a to the line defined by b and c
inline float distance(const point2& a, const point2& b, const point2& c) {
  const float l = (a-b).dot(c-b) / (c-b).length();
  const vec2 n = (c-b).normalized();
  const point2 d = b+l*n;
  return (a-d).length();
}

inline vec2 reflection(const vec2& v, const point2& a, const point2& b) {
  const vec2 n = (b-a).normalized();
  const float dot = v.dot(n);
  return 2*dot*n - v;
}

class bb2 {
 public:
  bb2(const float llx, const float lly,
      const float w, const float h)
      : ll(llx, lly), lr(llx+w, lly),
        ur(llx+w, lly+h), ul(llx, lly+h) {
  }
  bool in(const point2& p, const float buffer) const {
    return p.x() > ll.x() + buffer &&
        p.x() < ur.x() - buffer &&
        p.y() > ll.y() + buffer &&
        p.y() < ur.y() -buffer;
  }
  std::pair<point2, point2> closest_line(const point2& p) const {
    std::pair<point2, point2> closest;
    float min_dist = FLT_MAX;
    float dist;
    if ((dist = distance(p, ll, lr)) < min_dist) {
      min_dist = dist;
      closest = std::make_pair(ll, point2(ur.x(), ll.y()));
    }
    if ((dist = distance(p, lr, ur)) < min_dist) {
      min_dist = dist;
      closest = std::make_pair(point2(ur.x(), ll.y()), ur);
    }
    if ((dist = distance(p, ur, ul)) < min_dist) {
      min_dist = dist;
      closest = std::make_pair(ur, point2(ll.x(), ur.y()));
    }
    if ((dist = distance(p, ul, ll)) < min_dist) {
      min_dist = dist;
      closest = std::make_pair(point2(ll.x(), ur.y()), ll);
    }
    return closest;
  }

 private:
  point2 ll;
  point2 lr;
  point2 ur;
  point2 ul;
};

inline std::ostream& operator<<(std::ostream& out, const vec2& p) {
  out << "(" << p.x() << ", " << p.y() << ")";
  return out;
}

inline std::ostream& operator<<(std::ostream& out, const point2& p) {
  out << "(" << p.x() << ", " << p.y() << ")";
  return out;
}

#endif
